import argparse
import copy
import os
import re
import warnings
from typing import List, Optional

import pymeshlab
from lxml import etree
from pydrake.all import PackageMap


def _convert_mesh(
    url: str, path: str, scale: Optional[List[float]] = None, overwrite: bool = False
) -> str:
    if scale:
        scale_str = "_scaled_" + "_".join(
            [str(int(s) if s.is_integer() else s).replace("-", "n") for s in scale]
        )
    else:
        scale_str = ""
    suffix = "_from_" + path.rsplit(".", 1)[1] + scale_str + ".obj"

    output_url = url.rsplit(".", 1)[0] + suffix
    output_path = path.rsplit(".", 1)[0] + suffix

    if not overwrite and os.path.exists(output_path):
        # This was a bit too noisy:
        # print(f"Note: {output_path} already exists. Skipping conversion.")
        return output_url, output_path

    # Create a new PyMeshLab mesh set
    ms = pymeshlab.MeshSet()

    # Load the mesh file
    ms.load_new_mesh(path)

    if scale is not None:
        scale = [float(s) for s in scale]
        assert len(scale) == 3, "Scale must be a 3-element array."

        # First apply absolute value scaling
        ms.apply_filter(
            "compute_matrix_from_translation_rotation_scale",
            scalex=scale[0],
            scaley=scale[1],
            scalez=scale[2],
            freeze=True,
        )

    # Save the mesh
    ms.save_current_mesh(output_path, save_face_color=False, save_vertex_color=False)
    print(f"Converted {path} to {output_path}")
    return output_url, output_path


def _convert_urdf(input_filename, output_filename, package_map, overwrite):
    with open(input_filename, "r") as file:
        urdf_content = file.read()

    # Remove XML comments to avoid matching filenames inside them
    urdf_content_no_comments = re.sub(r"<!--.*?-->", "", urdf_content, flags=re.DOTALL)

    # Regex pattern to find entire XML nodes with .stl or .dae extensions in the filename attribute
    resource_pattern = re.compile(
        r'<(\w+)\s+[^>]*filename=["\']([^"\']+\.(stl|dae|obj))["\'][^>]*>',
        re.IGNORECASE,
    )

    # Find all matches in the URDF content without comments
    matches = resource_pattern.finditer(urdf_content_no_comments)

    modified_urdf_content = urdf_content

    for match in matches:
        node_text = match.group(0)
        node = etree.fromstring(node_text)
        mesh_url = node.attrib["filename"]
        scale = node.attrib.get("scale")
        if scale:
            scale = [float(s) for s in scale.split()]
            if len(set(scale)) == 1:
                # Uniform scaling is supported natively by Drake.
                scale = None
        if mesh_url.lower().endswith(".obj") and scale is None:
            # Don't need to convert .obj files with no scale or uniform scale.
            continue

        if mesh_url.lower().startswith("package://") or mesh_url.lower().startswith(
            "file://"
        ):
            mesh_path = package_map.ResolveUrl(mesh_url)
        else:
            mesh_path = os.path.join(os.path.dirname(input_filename), mesh_url)

        output_mesh_url, output_mesh_path = _convert_mesh(
            url=mesh_url, path=mesh_path, scale=scale, overwrite=overwrite
        )

        # Update the node's filename attribute
        node.attrib["filename"] = output_mesh_url

        # If scale is not None, update the scale attribute
        if scale:
            node.attrib["scale"] = "1 1 1"

        # Convert the node back to a string
        node_string = etree.tostring(node, encoding="unicode")

        # Replace the original match with the updated node string
        modified_urdf_content = modified_urdf_content.replace(node_text, node_string)

    with open(output_filename, "w") as file:
        file.write(modified_urdf_content)

    print(f"Converted URDF file '{input_filename}' to '{output_filename}'.")


def _convert_sdf(input_filename, output_filename, package_map, overwrite):
    raise NotImplementedError("SDF format is not supported yet, but it will be soon.")


def _process_includes(filename, processed_files=None):
    """Process include tags recursively and return complete DOM tree.

    Args:
        filename: Path to the input XML file
        processed_files: Set of already processed files to prevent cycles

    Returns:
        etree.ElementTree: Complete DOM tree with includes resolved
    """
    if processed_files is None:
        processed_files = set()

    if filename in processed_files:
        raise ValueError(
            f"Circular include detected: {filename} has already been processed"
        )

    processed_files.add(filename)

    # Parse the main file
    tree = etree.parse(filename)
    root = tree.getroot()

    # Find all include elements
    includes = root.findall(".//include")

    # Process each include
    for include in includes:
        # Get included file path relative to current file's directory
        include_file = include.get("file")
        print(f"Processing include file: {include_file}")
        if not include_file:
            raise ValueError("Include element missing required 'file' attribute")

        include_path = os.path.join(os.path.dirname(filename), include_file)

        # Process the included file recursively
        included_tree = _process_includes(include_path, processed_files)
        included_root = included_tree.getroot()

        # Insert all child elements from included file
        for child in included_root:
            # Insert before the include element
            include.addprevious(child)

        # Remove the include element
        include.getparent().remove(include)

    return tree


def _apply_defaults(element, defaults_dict, class_name_override=None):
    """Apply default attributes from a defaults class to an element.

    Args:
        element: The XML element to apply defaults to
        defaults_dict: Dictionary mapping (class_name, element_type) to default elements
    """
    if "class" in element.attrib:
        class_name = (
            element.attrib["class"]
            if class_name_override is None
            else class_name_override
        )
        # Try to find defaults for this specific element type and class
        key = (class_name, element.tag)
        if key in defaults_dict:
            for attr, value in defaults_dict[key].attrib.items():
                # Special case: only one rotation specification allowed
                rotation_attrs = {"quat", "axisangle", "xyaxes", "zaxis", "euler"}
                if attr in rotation_attrs:
                    # Check if element already has any rotation attributes
                    has_rotation = any(
                        rot_attr in element.attrib for rot_attr in rotation_attrs
                    )
                    if has_rotation:
                        continue
                if attr != "class":
                    if attr not in element.attrib:
                        element.attrib[attr] = value


def _convert_mjcf(input_filename, output_filename, package_map, overwrite):
    # Process includes first to build complete DOM
    tree = _process_includes(input_filename)
    root = tree.getroot()

    compiler_elements = root.findall(".//compiler")
    strippath = False
    meshdir = None
    for compiler_element in compiler_elements:
        if "strippath" in compiler_element.attrib:
            # Convert XML string "true"/"false" to Python bool
            strippath = compiler_element.attrib["strippath"].lower() == "true"
        if "assetdir" in compiler_element.attrib:
            meshdir = compiler_element.attrib["assetdir"]
            compiler_element.attrib["assetdir"]
        if "meshdir" in compiler_element.attrib:
            meshdir = compiler_element.attrib["meshdir"]
        if "texturedir" in compiler_element.attrib:
            compiler_element.attrib["texturedir"]

    # Dictionary to store default classes, initialize with empty "main" class
    # Initialize defaults with empty dictionary
    defaults = {}

    # Process all default elements recursively
    def process_defaults(element, parent_class="main"):
        """Process default elements recursively.

        Args:
            element: The default element to process
            parent_class: The parent class name for inheritance
        """
        # Get class name, default to parent_class if not specified
        class_name = element.get("class", parent_class)

        # Process this default element's children (which define defaults for specific types)
        for child in element:
            if child.tag != "default":  # Only process non-default children
                # Create key from class name and element type
                key = (class_name, child.tag)

                # If parent class has defaults for this type, inherit them
                parent_key = (parent_class, child.tag)
                if parent_key in defaults:
                    defaults[key] = copy.deepcopy(defaults[parent_key])
                else:
                    # Create new default element
                    defaults[key] = etree.Element(child.tag)

                # Update with new attributes
                for attr, value in child.attrib.items():
                    if attr != "class":
                        defaults[key].attrib[attr] = value

        # Process nested default elements recursively
        for child in element.findall("default"):
            process_defaults(child, class_name)

    # Find and process all top-level default elements
    for default_element in root.findall("default"):
        process_defaults(default_element)

    # Find all <mesh> elements recursively using xpath
    mesh_elements = root.findall(".//asset/mesh")
    for mesh_element in mesh_elements:
        _apply_defaults(mesh_element, defaults)

        mesh_url = mesh_element.attrib["file"]
        if mesh_url is None:
            raise ValueError("A 'file' attribute must be specified for mesh elements.")
        scale = mesh_element.attrib.get("scale")
        if scale:
            scale = [float(s) for s in scale.split()]
            if len(set(scale)) == 1:
                # Uniform scaling is supported natively by Drake.
                scale = None
        if mesh_url.lower().endswith(".obj") and scale is None:
            # Don't need to convert .obj files with no scale or uniform scale.
            continue

        # Get absolute path to mesh file according to MJCF rules
        if strippath:
            # Remove all path information, keeping only filename
            mesh_url = os.path.basename(mesh_url)

        # Check if mesh_url is already an absolute path
        if os.path.isabs(mesh_url):
            mesh_path = mesh_url
        # Check if meshdir is an absolute path
        elif meshdir is not None and os.path.isabs(meshdir):
            mesh_path = os.path.join(meshdir, mesh_url)
        # Use path relative to MJCF file
        else:
            base_path = os.path.dirname(input_filename)
            if meshdir is not None:
                base_path = os.path.join(base_path, meshdir)
            mesh_path = os.path.join(base_path, mesh_url)

        if not os.path.exists(mesh_path):
            if meshdir is not None:
                warnings.warn(
                    f"The file {mesh_path} does not exist in the expected location. This may be due to the fact that meshdir and assetdir are not being parsed yet.",
                    UserWarning,
                )
            raise FileNotFoundError(f"The file {mesh_path} does not exist.")

        output_mesh_url, output_mesh_path = _convert_mesh(
            url=mesh_url, path=mesh_path, scale=scale, overwrite=overwrite
        )

        # Update the node's filename attribute
        mesh_element.attrib["file"] = output_mesh_url

        # If scale is not None, update the scale attribute
        if scale:
            mesh_element.attrib["scale"] = "1 1 1"

    geoms_with_type_plane = root.findall(".//geom[@type='plane']")
    for geom in geoms_with_type_plane:
        parent = geom.getparent()
        while parent is not None:
            if parent.tag == "worldbody":
                break
            elif parent.tag == "body":
                # Then the plane would have been parsed as a dynamic collision element in Drake. Replace it with a large box.
                geom.attrib["size"] = "1000 1000 1"
                if "pos" in geom.attrib:
                    pos = [float(value) for value in geom.attrib["pos"].split()]
                else:
                    pos = [0, 0, 0]
                pos[2] -= 1
                geom.attrib["pos"] = " ".join([str(value) for value in pos])
                geom.attrib["type"] = "box"
                break
            parent = parent.getparent()

    tree.write(output_filename, pretty_print=True)
    print(f"Converted MJCF file '{input_filename}' to '{output_filename}'.")


def MakeDrakeCompatibleModel(
    input_filename: str,
    output_filename: str,
    package_map: PackageMap = None,
    overwrite: bool = False,
) -> None:
    """Converts a model file (currently .urdf or .xml)to be compatible with the
    Drake multibody parsers.

    - Converts any .stl files to obj
      https://github.com/RobotLocomotion/drake/issues/19408
    - Converts any .dae files to obj
      https://github.com/RobotLocomotion/drake/issues/19109
    - Zaps any non-uniform scale attributes
      https://github.com/RobotLocomotion/drake/issues/22046
    - Converts dynamic half-space collision geometries to (very) large boxes
      https://github.com/RobotLocomotion/drake/issues/19263
      (this is so far implemented only for mujoco .xml models)

    Any new files will be created alongside the original files (e.g. .obj files
    will be created next to the existing .stl files); all new files will get a
    descriptive suffix, and existing files will not be overwritten by default.

    Args:
        input_filename (str): The path to the input file to be converted.
        output_filename (str): The path where the converted file will be saved.
            Using the same string as input_filename is allowed.
        package_map (PackageMap, optional): The package map to use. Defaults to None.
        overwrite (bool, optional): Whether to overwrite existing files. Defaults
            to False.
    """
    if package_map is None:
        package_map = PackageMap()

    if input_filename.lower().endswith(".urdf"):
        _convert_urdf(input_filename, output_filename, package_map, overwrite=overwrite)
    elif input_filename.lower().endswith(".sdf"):
        _convert_sdf(input_filename, output_filename, package_map, overwrite=overwrite)
    elif input_filename.lower().endswith(".xml"):
        _convert_mjcf(input_filename, output_filename, package_map, overwrite=overwrite)
    else:
        print(
            f"Warning: The file extension of '{input_filename}' is not "
            "supported yet; we currently support '.stl', '.dae', and '.urdf'."
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Convert STL, DAE, or URDF files to use OBJ instead of STL or DAE."
    )
    parser.add_argument(
        "input_filename", type=str, help="The path to the input file to be converted."
    )
    parser.add_argument(
        "output_filename",
        type=str,
        help="The path where the converted file will be saved.",
    )
    parser.add_argument(
        "--populate_package_map_from_env",
        type=str,
        default=None,
        help="Environment variable to populate the package map from.",
    )
    parser.add_argument(
        "--populate_package_map_from_folder",
        type=str,
        default=None,
        help="Folder path to populate the package map from.",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Whether to overwrite existing files.",
    )

    args = parser.parse_args()

    package_map = PackageMap()

    if args.populate_package_map_from_env:
        package_map.PopulateFromEnvironment(args.populate_package_map_from_env)

    if args.populate_package_map_from_folder:
        package_map.PopulateFromFolder(args.populate_package_map_from_folder)

    MakeDrakeCompatibleModel(
        input_filename=args.input_filename,
        output_filename=args.output_filename,
        package_map=package_map,
        overwrite=args.overwrite,
    )



import unittest

from pydrake.all import StartMeshcat

from manipulation.station import LoadScenario, MakeHardwareStation, Scenario


class HardwareStationInterfaceTest(unittest.TestCase):
    def get_scenario(self) -> Scenario:
        scenario_data = """
directives:
    # Add iiwa_left
    - add_model:
        name: iiwa_left
        file: package://drake_models/iiwa_description/urdf/iiwa14_spheres_collision.urdf

    - add_weld:
        parent: world
        child: iiwa_left::base

    # Add schunk_left
    - add_model:
        name: wsg_left
        file: package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf

    - add_weld:
        parent: iiwa_left::iiwa_link_7
        child: wsg_left::body
        X_PC:
            translation: [0, 0, 0.114]
            rotation: !Rpy { deg: [90.0, 0.0, 68.0 ]}

    # Add iiwa_right
    - add_model:
        name: iiwa_right
        file: package://drake_models/iiwa_description/urdf/iiwa14_no_collision.urdf

    - add_weld:
        parent: world
        child: iiwa_right::base
        X_PC:
            translation: [0, 0.765, 0]

    # Add schunk_right
    - add_model:
        name: wsg_right
        file: package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf

    - add_weld:
        parent: iiwa_right::iiwa_link_7
        child: wsg_right::body
        X_PC:
            translation: [0, 0, 0.114]
            rotation: !Rpy { deg: [90.0, 0.0, 68.0 ]}

    # Add table
    - add_model:
        name: table
        file: package://drake_models/manipulation_station/table_wide.sdf

    - add_frame:
        name: table_origin
        X_PF:
            base_frame: world
            translation: [0.4, 0.3825, 0.0]
            rotation: !Rpy { deg: [0., 0., 0.]}

    - add_weld:
        parent: table_origin
        child: table::table_body

lcm_buses:
    left_lcm:
        channel_suffix: _LEFT
    right_lcm:
        channel_suffix: _RIGHT

model_drivers:
    iiwa_left: !IiwaDriver
        control_mode: position_only
        hand_model_name: wsg_left
        lcm_bus: left_lcm
    iiwa_right: !IiwaDriver
        control_mode: position_only
        hand_model_name: wsg_right
        lcm_bus: right_lcm
    wsg_left: !SchunkWsgDriver {}
    wsg_right: !SchunkWsgDriver {}
"""
        return LoadScenario(data=scenario_data)

    def test_without_meshcat(self):
        scenario = self.get_scenario()
        station = MakeHardwareStation(scenario, hardware=True, meshcat=None)

        # Should not contain SceneGraph and MeshcatVisualizer.
        self.assertFalse(station.HasSubsystemNamed("scene_graph"))
        self.assertFalse(station.HasSubsystemNamed("meshcat_visualizer(illustration)"))
        self.assertFalse(station.HasSubsystemNamed("meshcat_visualizer(proximity)"))

    def test_with_meshcat(self):
        scenario = self.get_scenario()
        meshcat = StartMeshcat()
        station = MakeHardwareStation(scenario, hardware=True, meshcat=meshcat)

        # Should contain SceneGraph and MeshcatVisualizer.
        self.assertTrue(station.HasSubsystemNamed("scene_graph"))
        self.assertTrue(station.HasSubsystemNamed("meshcat_visualizer(illustration)"))
        self.assertTrue(station.HasSubsystemNamed("meshcat_visualizer(proximity)"))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight

tri_obs = np.array(
    [
        [
            [0.98886109, 0.81957125],
            [0.92702203, 0.71246285],
            [1.05070015, 0.71246285],
        ],
        [
            [0.28044399, 0.86069726],
            [0.21859425, 0.75357036],
            [0.34229373, 0.75357036],
        ],
        [
            [0.10322601, 0.53148317],
            [0.03083525, 0.40609871],
            [0.17561676, 0.40609871],
        ],
        [
            [0.9085955, 0.37510306],
            [0.83802404, 0.25286969],
            [0.97916697, 0.25286969],
        ],
        [
            [0.28777534, 0.21868408],
            [0.21099742, 0.08570082],
            [0.36455326, 0.08570082],
        ],
        [
            [0.01936696, 0.722583],
            [-0.01851946, 0.6569618],
            [0.05725338, 0.6569618],
        ],
        [
            [0.21162812, 0.37628579],
            [0.11572522, 0.21017709],
            [0.30753101, 0.21017709],
        ],
        [
            [0.49157316, 0.09989796],
            [0.4512723, 0.03009484],
            [0.53187401, 0.03009484],
        ],
        [
            [0.57411761, 0.21150516],
            [0.51801944, 0.11434028],
            [0.63021577, 0.11434028],
        ],
        [
            [0.58930554, 0.80235816],
            [0.5004515, 0.64845846],
            [0.67815957, 0.64845846],
        ],
        [[0.0, 0.0], [1.0, 0.0], [0.5, -0.5]],
        [[1.0, 0.0], [1.0, 1.0], [1.5, 0.5]],
        [[1.0, 1.0], [0.0, 1.0], [0.5, 1.5]],
        [[0.0, 1.0], [0.0, 0.0], [-0.5, 0.5]],
    ]
)


class TestTaskspaceIRIS(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(5.0)
    def test_closest_point_ellipse(self):
        """Test closest point"""
        f = self.notebook_locals["ClosestPointOnObstacle"]

        test_C = np.array([[0.22810313, -0.10359835], [-0.0444466, 0.27173404]])
        test_C_inv = np.linalg.inv(test_C)
        test_d = np.array([0.36327963, 0.4762024])

        test_idxs = [7, 12, 1, 9]

        x_star_sol = np.array(
            [
                [0.49157316, 0.09989796],
                [0.09874051, 1.0],
                [0.22319751, 0.75357036],
                [0.5004515, 0.64845846],
            ]
        )
        dist_sol = np.array(
            [
                1.3984260160646136,
                1.902331715469969,
                1.0073468778009206925,
                1.24440873680021546,
            ]
        )

        tris = [val for val in tri_obs]

        x_diff = []
        dist_diff = []
        for i, idx in enumerate(test_idxs):
            o = tris[idx]
            x_star_pred, dist_pred = f(test_C, test_C_inv, test_d, o)

            x_diff.append(np.linalg.norm(x_star_sol[i] - x_star_pred))
            dist_diff.append(np.linalg.norm(dist_sol[i] - dist_pred))

        x_diff = np.asarray(x_diff)
        dist_diff = np.asarray(dist_diff)

        self.assertTrue((x_diff < 1e-3).all(), "Closest points are wrong!")
        self.assertTrue((dist_diff < 1e-2).all(), "Closest distances are wrong!")

    @weight(4)
    @timeout_decorator.timeout(5.0)
    def test_separating_hyperplane(self):
        """Test separating hyperplane"""
        f = self.notebook_locals["SeparatingHyperplanes"]

        tris = [val for val in tri_obs]
        test_C = np.array([[0.22810313, -0.10359835], [-0.0444466, 0.27173404]])
        test_d = np.array([0.36327963, 0.4762024])
        A, b, found = f(test_C, test_d, tris)

        if False:  # For debugging
            import matplotlib.pyplot as plt
            from matplotlib.patches import Polygon

            fig, ax = plt.subplots()
            draw_intersection = self.notebook_locals["draw_intersection"]
            draw_intersection(A, b, test_d, ax)
            for region in tri_obs:
                plt.plot(region[:, 0], region[:, 1], color="green", alpha=0.5)
                plt.plot(
                    region[[0, -1], 0], region[[0, -1], 1], color="green", alpha=0.5
                )
                ax.add_patch(Polygon(region, color="green", alpha=0.25))
            plt.show()

        self.assertTrue(found, "No solution found for hyperplanes!")

        self.assertEqual(A.shape, (2, 8), "We expect 8 hyperplanes")
        self.assertEqual(b.shape, (8, 1), "We expect 8 hyperplanes")

        # Check that the point is on one side of all the hyperplanes
        self.assertTrue(
            (A.T @ test_d < b.T).all(),
            "The center point, d, of the ellipse is outside the resulting polytope.",
        )

        # Check each obstacles is outside the at least one hyperplane
        epsilon = 1e-3
        for obs in tri_obs:
            all_vertices_outside = np.all(A.T @ obs.T >= b - epsilon, axis=1)
            at_least_one_separating_hyperplane = np.any(all_vertices_outside)
            self.assertTrue(
                at_least_one_separating_hyperplane,
                f"An obstacle is inside the resulting polytope.",
            )



import unittest

import numpy as np
from pydrake.all import Simulator

from manipulation.station import LoadScenario, MakeHardwareStation


class InverseDynamicsDriverTest(unittest.TestCase):
    def get_bimanual_directives_data(self):
        return """
directives:
- add_model:
    name: iiwa_left
    file: package://drake_models/iiwa_description/urdf/iiwa14_no_collision.urdf
- add_frame:
    name: iiwa_left_origin
    X_PF:
        base_frame: world
        translation: [0, 0.765, 0]
- add_weld:
    parent: iiwa_left_origin
    child: iiwa_left::base
- add_model:
    name: wsg_left
    file: package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf
- add_frame:
    name: iiwa_left::wsg_attach
    X_PF:
        base_frame: iiwa_left::iiwa_link_7
        translation: [0, 0, 0.114]
        rotation: !Rpy { deg: [90, 0, 68]}
- add_weld:
    parent: iiwa_left::iiwa_link_7
    child: wsg_left::body
    X_PC:
        translation: [0, 0, 0.09]
        rotation: !Rpy { deg: [90, 0, 90]}
- add_model:
    name: iiwa_right
    file: package://drake_models/iiwa_description/urdf/iiwa14_no_collision.urdf
- add_frame:
    name: iiwa_right_origin
    X_PF:
        base_frame: world
        translation: [0, -0.765, 0]
- add_weld:
    parent: iiwa_right_origin
    child: iiwa_right::base
- add_model:
    name: wsg_right
    file: package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf
- add_frame:
    name: iiwa_right::wsg_attach
    X_PF:
        base_frame: iiwa_right::iiwa_link_7
        translation: [0, 0, 0.114]
        rotation: !Rpy { deg: [90, 0, 68]}
- add_weld:
    parent: iiwa_right::iiwa_link_7
    child: wsg_right::body
    X_PC:
        translation: [0, 0, 0.09]
        rotation: !Rpy { deg: [90, 0, 90]}
        """

    def make_controller_test(self, model_name: str, model_dof: int):
        directives_data = self.get_bimanual_directives_data()
        model_drivers_data = f"""
model_drivers:
    {model_name}: !InverseDynamicsDriver {{}}
        """
        scenario = LoadScenario(data=directives_data + model_drivers_data)
        station = MakeHardwareStation(scenario)

        simulator = Simulator(station)
        context = simulator.get_mutable_context()
        fixed_value = np.zeros(2 * model_dof)
        station.GetInputPort(f"{model_name}.desired_state").FixValue(
            context, fixed_value
        )
        simulator.AdvanceTo(1.0)

    def test_controller_on_single_model(self):
        self.make_controller_test(
            model_name="iiwa_left",
            model_dof=7,
        )

    def test_controller_on_two_models(self):
        self.make_controller_test(
            model_name="iiwa_left+wsg_left",
            model_dof=9,
        )

    def test_controller_on_three_models(self):
        self.make_controller_test(
            model_name="iiwa_left+wsg_left+iiwa_right",
            model_dof=16,
        )

    def test_controller_on_all_models(self):
        self.make_controller_test(
            model_name="iiwa_left+iiwa_right+wsg_left+wsg_right",
            model_dof=18,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import pydrake.multibody.parsing
import pydrake.multibody.plant
import pydrake.systems.framework

from manipulation.utils import ConfigureParser


class TestModelDirectives(unittest.TestCase):
    def test_model_loading(self):
        filename = "package://manipulation/two_bins_w_cameras.dmd.yaml"

        builder = pydrake.systems.framework.DiagramBuilder()
        plant, scene_graph = pydrake.multibody.plant.AddMultibodyPlantSceneGraph(
            builder, time_step=0.001
        )
        parser = pydrake.multibody.parsing.Parser(plant)
        ConfigureParser(parser)
        parser.AddModelsFromUrl(filename)
        plant.Finalize()

        diagram = builder.Build()
        context = diagram.CreateDefaultContext()
        # Test that we can create a context without errors
        self.assertIsNotNone(context)


if __name__ == "__main__":
    unittest.main()



# Written by Caelan Garrett, modified a bit by Tomas Lozano-Perez
# yapf: disable
from math import cos, pi, sin
from random import randint, random, uniform
from time import time

from .geometry import AABB, Line, Object, Point, Polygon, convex_hull
from .robot import ConfigurationSpace


class TreeNode:

    def __init__(self, value, parent=None):
        self.value = value  # tuple of floats representing a configuration
        self.parent = parent  # another TreeNode
        self.children = []  # list of TreeNodes


class RRT:
    """RRT Tree."""

    def __init__(self, root: TreeNode, cspace: ConfigurationSpace):
        self.root = root  # root TreeNode
        self.cspace = cspace  # robot.ConfigurationSpace
        self.size = 1  # int length of path
        self.max_recursion = 1000  # int length of longest possible path

    def add_configuration(self, parent_node, child_value):
        child_node = TreeNode(child_value, parent_node)
        parent_node.children.append(child_node)
        self.size += 1
        return child_node

    # Brute force nearest, handles general distance functions
    def nearest(self, configuration):
        """
        Finds the nearest node by distance to configuration in the
        configuration space.

        Args:
            configuration: tuple of floats representing a configuration of a
                robot

        Returns:
            closest: TreeNode. the closest node in the configuration space
                to configuration
            distance: float. distance from configuration to closest
        """
        assert self.cspace.valid_configuration(configuration)

        def recur(node, depth=0):
            closest, distance = node, self.cspace.distance(
                node.value, configuration)
            if depth < self.max_recursion:
                for child in node.children:
                    (child_closest, child_distance) = recur(child, depth + 1)
                    if child_distance < distance:
                        closest = child_closest
                        distance = child_distance
            return closest, distance

        return recur(self.root)[0]

    def draw(self, window, color='black'):

        def recur(node, depth=0):
            node_point = Point(node.value[0], node.value[1])
            if depth < self.max_recursion:
                for child in node.children:
                    child_point = Point(child.value[0], child.value[1])
                    Line(node_point, child_point).draw(window,
                                                       color=color,
                                                       width=1)
                    recur(child, depth + 1)

        recur(self.root)


class Problem:
    """Defines a motion-planning problem."""

    def __init__(self,
                 x,
                 y,
                 robot,
                 obstacles,
                 start,
                 goal,
                 cspace,
                 display_tree=False):
        """
        Defines a motion planning problem.

        Args:
            x: float. the width of the map's area
            y: float. the height of the map's area
            robot: a robot.Robot instance
            obstacles: list of geometry.Objects self.robot can't move through
            start: tuple of floats: starting configuration of self.robot
            goal: tuple of floats: goal configuration of self.robot
            cspace: robot.Configuration space of self.robot
            display_tree: bool. if True, draw the generated plan trees
        """
        self.x = x
        self.y = y
        self.robot = robot
        self.obstacles = obstacles
        self.start = start
        self.goal = goal
        self.region = AABB(Point(0, 0), Point(x, y))
        self.cspace = cspace
        self.display_tree = display_tree

        assert self.valid_configuration(self.start)
        assert self.valid_configuration(self.goal)

    # Generate a regular polygon that does not collide with other
    # obstacles or the robot at start and goal.
    def generate_random_regular_poly(self,
                                     num_verts,
                                     radius,
                                     angle=uniform(-pi, pi)):
        """
        Generates a regular polygon that does not collide with other
        obstacles or the robot at start and goal. This polygon is added to
        self.obstacles. To make it random, keep the default angle argument.

        Args:
            num_verts: int. the number of vertices of the polygon >= 3
            radius: float. the distance from the center of the polygon
                to any vertex > 0
            angle: float. the angle in radians between the origin and
                the first vertex. the default is a random value between
                -pi and pi.
        """
        (min_verts, max_verts) = num_verts
        (min_radius, max_radius) = radius
        assert not (min_verts < 3 or min_verts > max_verts or min_radius <= 0
               or min_radius > max_radius)
        reference = Point(random() * self.x, random() * self.y)
        distance = uniform(min_radius, max_radius)
        sides = randint(min_verts, max_verts)
        obj = Object(reference, [
            Polygon([
                Point(distance * cos(angle + 2 * n * pi / sides),
                      distance * sin(angle + 2 * n * pi / sides))
                for n in range(sides)
            ])
        ])
        if any([obj.collides(current) for current in self.obstacles]) or \
                obj.collides(self.robot.configuration(self.start)) or \
                obj.collides(self.robot.configuration(self.goal)) or \
                not self.region.contains(obj):
            self.generate_random_regular_poly(num_verts, radius, angle=angle)
        else:
            self.obstacles.append(obj)

    # Generate a polygon that does not collide with other
    # obstacles or the robot at start and goal.
    def generate_random_poly(self, num_verts, radius):
        """
        Generates a random polygon that does not collide with other
        obstacles or the robot at start and goal. This polygon is added to
        self.obstacles.

        Args:
            num_verts: int. the number of vertices of the polygon >= 3
            radius: float. a reference distance between the origin and some
                vertex of the polygon > 0
        """
        (min_verts, max_verts) = num_verts
        (min_radius, max_radius) = radius
        assert not (min_verts < 3 or min_verts > max_verts or min_radius <= 0
               or min_radius > max_radius)

        reference = Point(random() * self.x, random() * self.y)
        verts = randint(min_verts, max_verts)
        points = [Point(0, 0)]
        for i in range(verts):
            angle = 2 * pi * random()
            points.append(((max_radius - min_radius) * random() + min_radius)
                          * Point(cos(angle), sin(angle)))
        obj = Object(reference, [Polygon(convex_hull(points))])
        if any([obj.collides(current) for current in self.obstacles]) or \
                obj.collides(self.robot.configuration(self.start)) or \
                obj.collides(self.robot.configuration(self.goal)) or \
                not self.region.contains(obj):
            self.generate_random_poly(num_verts, radius)
        else:
            self.obstacles.append(obj)

    def valid_configuration(self, configuration):
        """
        Checks if the given configuration is valid in this Problem's
        configuration space and doesn't collide with any obstacles.

        Args:
            configuration: tuple of floats - tuple describing a robot
                configuration

        Returns:
            bool. True if the given configuration is valid, False otherwise
        """
        return self.cspace.valid_configuration(configuration) \
            and not self.collide(configuration)

    def collide(self, configuration):  # Check collision for configuration
        """
        Checks if the given configuration collides with any of this
        Problem's obstacles.

        Args:
            configuration: tuple of floats - tuple describing a robot
                configuration

        Returns:
            bool. True if the given configuration is in collision,
            False otherwise
        """
        config_robot = self.robot.configuration(configuration)
        return any([config_robot.collides(obstacle)
                    for obstacle in self.obstacles]) or \
            not self.region.contains(config_robot)

    def test_path(self, start, end):  # check path for collisions
        """
        Checks if the path from start to end collides with any obstacles.

        Args:

            start: tuple of floats - tuple describing robot's start
                configuration
            end: tuple of floats - tuple describing robot's end configuration

        Returns:
            bool. False if the path collides with any obstacles, True otherwise
        """
        path = self.cspace.path(start, end)
        for configuration in path:
            if self.collide(configuration):
                return False
        return True

    def safe_path(self, start, end):  # returns subset of path that is safe
        """
        Checks if the path from start to end collides with any obstacles.

        Args:
            start: tuple of floats - tuple describing robot's start
                configuration
            end: tuple of floats - tuple describing robot's end configuration

        Returns:
            list of tuples along the path that are not in collision.
        """
        path = self.cspace.path(start, end)
        safe_path = []
        for configuration in path:
            if self.collide(configuration):
                return safe_path
            safe_path.append(configuration)
        return safe_path

    def path_distance(self, path):  # add up distances along path
        """
        Adds up the distance along the given path.

        Args:
            path: list of tuples describing configurations of the robot

        Returns:
            float. the total distance along the path
        """
        distance = 0
        for i in range(len(path) - 1):
            distance += self.cspace.distance(path[i], path[i + 1])
        return distance

    # Given a path (list of configurations) return a (shorter) path
    def smooth_path(self, path, attempts=100):  # random short-cutting
        """
        Given a path (list of configurations), return a possibly shorter
        path.

        Args:

            path: list of tuples describing configurations of the robot
            attemps: int. the number of times to try to smooth the path

        Returns:
            list of tuples describing configurations of the robot that is
            possibly shorter than the given path
        """
        #########################
        # Your code here

        smoothed_path = path
        for attempt in range(attempts):
            if len(smoothed_path) <= 2:
                return smoothed_path
            i = randint(0, len(smoothed_path) - 1)
            j = randint(0, len(smoothed_path) - 1)
            if i == j or abs(i - j) == 1:
                continue
            one, two = i, j
            if j < i:
                one, two = j, i
            if self.test_path(smoothed_path[one], smoothed_path[two]):
                smoothed_path = smoothed_path[:one + 1] + smoothed_path[two:]
        return smoothed_path

        #########################
        raise NotImplemented

    def draw_robot_path(self, path, color=None):
        """
        Draws the robot in each configuration along a path in a Tkinter
        window.

        Args:
            path: list of tuples describing configurations
                of the robot
            color: string. Tkinter color of the robot
        """
        for i in range(1, len(path) - 1):  # don't draw start and end
            self.robot.configuration(path[i]).draw(self.window, color)
            self.window.update()
            input('Next?')

    def run_and_display(self, method, display=True):
        """
        Runs the Problem with the given methods and draws the path in a
        Tkinter window.

        Args:
            method: list of strings of the methods to try, must be "rrt" or
                "birrt"
            display: bool. if True, draw the Problem in a Tkinter window

        Returns:
            bool. True if a collision free path is found, False otherwise
        """

        def draw_problem():
            for obs in self.obstacles:
                obs.draw(self.window, 'red')
            self.robot.configuration(self.start).draw(self.window, 'orange')
            self.robot.configuration(self.goal).draw(self.window, 'green')

        # if display:
        #     self.window = DrawingWindow(600, 600, 0, self.x, 0, self.y,
        #     'RRT Planning')
        #     draw_problem()
        t1 = time()
        if method == 'rrt':
            path = self.rrt_planning()
        elif method == 'birrt':
            path = self.bidirectional_rrt_planning()
        else:
            raise NotImplemented
        print('RRT took', time() - t1, 'seconds')
        if display and self.display_tree:
            for rrt in self.rrts:
                rrt.draw(self.window)

        if path is None:
            print('No path found')
            return False
        else:
            print(
                'Path found with ' + str(len(path) - 1)
                + ' movements of distance ', self.path_distance(path))
            smooth_path = self.smooth_path(path)
            print(
                'Smoothed path found with ' + str(len(smooth_path) - 1)
                + ' movements of distance ', self.path_distance(smooth_path))
            # interpolated smooth path
            spath = []
            for i in range(1, len(smooth_path)):
                spath.extend(
                    self.cspace.path(smooth_path[i - 1], smooth_path[i]))
            # make sure path is collision free
            if any([self.collide(c) for c in spath]):
                print('Collision in smoothed path')
                return False
            if display:
                self.draw_robot_path(path, color='yellow')
                self.window.clear()
                draw_problem()
                self.draw_robot_path(spath, color='gold')
                while input('End? (y or n)') != 'y':
                    pass
        return True



import unittest

from pydrake.all import (
    DiagramBuilder,
    LcmImageArrayToImages,
    LcmSubscriberSystem,
    StartMeshcat,
)

from manipulation.station import LoadScenario, MakeHardwareStation


class StationCameraIdsTest(unittest.TestCase):
    def test_load_scenario_with_camera_ids(self):
        scenario_data = """
directives:
- add_frame:
    name: camera0_origin
    X_PF:
        base_frame: world
        rotation: !Rpy { deg: [-110, 0, 90]}
        translation: [1.2, 0.0, 0.54]

- add_model:
    name: camera_main
    file: package://manipulation/camera_box.sdf
cameras:
    main_camera:
        name: camera0
        lcm_bus: default
        depth: True
        X_PB:
            base_frame: camera_main::base
camera_ids:
    main_camera: DRAKE_RGBD_CAMERA_IMAGES_810512062206
        """

        meshcat = StartMeshcat()
        meshcat.ResetRenderMode()
        builder = DiagramBuilder()
        scenario = LoadScenario(data=scenario_data)
        station = builder.AddSystem(
            MakeHardwareStation(scenario, meshcat, hardware=True)
        )

        self.assertTrue(station.HasSubsystemNamed("camera0.data_subscriber"))
        self.assertTrue(station.HasSubsystemNamed("camera0.data_receiver"))

        in_system, out_system = next(iter(station.connection_map().items()))
        self.assertIsInstance(in_system[0], LcmImageArrayToImages)
        self.assertIsInstance(out_system[0], LcmSubscriberSystem)


if __name__ == "__main__":
    unittest.main()
